"""
Provides classes for handler representation
"""
from typing import List, Optional


class DetectPolicy:
    """
    Class describes detect Policy data structure

    Attributes:
        estimateHeadPose: estimated head pose
        estimateEmotions: whether to estimate emotions from the image
        estimateQuality: estimate head pose
        estimateGaze: whether to estimate eye gaze from the image.
        estimateEyesAttributes: whether to estimate eye attributes from the image
        estimateMouthAttributes: whether to estimate mouth attributes from the image
        detectLandmarks68: whether to detect basic 68-point facial landmarks on the image
        extractExif: whether to extract EXIF meta information from input JPEG images.
                     Exact output will vary since there are no mandatory requirements to both authoring
                     software and digital cameras how to write the data.
                     This function will only parse the tags and output their names and values as is.
                     Please refer to JEITA CP-3451 EXIF specification for details.
        yawThreshold: maximum deviation yaw angle from 0
        rollThreshold: maximum deviation roll angle from 0
        pitchThreshold: maximum deviation pitch angle from 0
    """

    def __init__(
        self,
        estimateHeadPose: int = 0,
        estimateEmotions: int = 0,
        estimateQuality: int = 0,
        estimateGaze: int = 0,
        estimateEyesAttributes: int = 0,
        estimateMouthAttributes: int = 0,
        detectLandmarks68: int = 0,
        extractExif: int = 0,
        yawThreshold: Optional[int] = None,
        rollThreshold: Optional[int] = None,
        pitchThreshold: Optional[int] = None,
    ):
        self.estimateHeadPose = estimateHeadPose
        self.estimateEmotions = estimateEmotions
        self.estimateQuality = estimateQuality
        self.estimateGaze = estimateGaze
        self.estimateEyesAttributes = estimateEyesAttributes
        self.estimateMouthAttributes = estimateMouthAttributes
        self.detectLandmarks68 = detectLandmarks68
        self.extractExif = extractExif
        self.yawThreshold = yawThreshold
        self.rollThreshold = rollThreshold
        self.pitchThreshold = pitchThreshold

    def asDict(self) -> dict:
        """
        Creates a dict with all existing values of this data structure

        Returns: dict with values of this data structure
        """
        allArgs = {
            "estimate_head_pose": self.estimateHeadPose,
            "estimate_emotions": self.estimateEmotions,
            "estimate_quality": self.estimateQuality,
            "estimate_gaze": self.estimateGaze,
            "estimate_eyes_attributes": self.estimateEyesAttributes,
            "estimate_mouth_attributes": self.estimateMouthAttributes,
            "detect_landmarks68": self.detectLandmarks68,
            "extract_exif": self.extractExif,
        }
        if self.yawThreshold is not None:
            allArgs["yaw_threshold"] = self.yawThreshold
        if self.rollThreshold is not None:
            allArgs["roll_threshold"] = self.rollThreshold
        if self.pitchThreshold is not None:
            allArgs["pitch_threshold"] = self.pitchThreshold
        return allArgs


class ExtractPolicy:
    """
    Class describes extract Policy data structure

    Attributes:
        extractBasicAttributes: flag that tells whether to extract basic attributes (gender, age, ethnicity)
        extractDescriptor: flag that tells whether to extract face descriptor(s)
        scoreThreshold: descriptor quality score threshold. All the attributes with quality score below the threshold
                        will be ignored (and not stored in the DB). The function will proceed as usual with all
                        the remaining descriptors (if left).
    """

    def __init__(self, extractBasicAttributes: int = 0, extractDescriptor: int = 1, scoreThreshold: float = 0.0):
        self.extractBasicAttributes = extractBasicAttributes
        self.extractDescriptor = extractDescriptor
        self.scoreThreshold = scoreThreshold

    def asDict(self) -> dict:
        """
        Creates a dict with all existing values of this data structure

        Returns: dict with values of this data structure
        """

        return {
            "extract_basic_attributes": self.extractBasicAttributes,
            "extract_descriptor": self.extractDescriptor,
            "score_threshold": self.scoreThreshold,
        }


class MatchPolicy:
    """
    Class that describes matching conditions (list + some of basic attributes).
    Every event that satisfies the filters for the list will be matched by it

    Attributes:
        listId: list id
        gender: Gender. 1 - male, 0 - female.
        ethnicities: Array of integer (ethnicities_filter) (Items Enum:1 2 3 4)
        ageLt: [ 0 .. 100 ] - upper bound for reference age
        ageGte: [ 0 .. 100 ] - lower including bound for reference age
    """

    def __init__(
        self,
        listId: str,
        gender: Optional[int] = None,
        ethnicities: Optional[List[int]] = None,
        ageLt: Optional[int] = None,
        ageGte: Optional[int] = None,
        limit: Optional[int] = None,
    ):
        self.listId = listId
        self.gender = gender
        self.ethnicities = ethnicities
        self.ageLt = ageLt
        self.ageGte = ageGte
        self.limit = limit

    def asDict(self) -> dict:
        """
        Creates a dict with all existing values of this data structure

        Returns: dict with values of this data structure
        """
        allArgs = {"list_id": self.listId, "filters": {}}
        if self.gender is not None:
            allArgs["filters"]["gender"] = self.gender
        if self.ethnicities is not None:
            allArgs["filters"]["ethnicities"] = self.ethnicities
        if self.ageLt is not None:
            allArgs["filters"]["age__lt"] = self.ageLt
        if self.ageGte is not None:
            allArgs["filters"]["age__gte"] = self.ageGte
        if self.limit is not None:
            allArgs["limit"] = self.limit
        return allArgs


class MatchFilter:
    """
    Describes a list for matching with optional similarity limitation

    Attributes:
        listId: id of a luna list which took part in the match
        similartityLte: [ 1 .. 0 ] Upper including bound for best match by the list
        similartityGte: [ 1 .. 0 ] Lower including bound for best match by the list
    """

    def __init__(
        self, listId: Optional[str] = None, similarityLte: Optional[float] = None, similarityGte: Optional[float] = None
    ):
        self.listId = listId
        self.similartityLte = similarityLte
        self.similartityGte = similarityGte

    def asDict(self) -> dict:
        """
        Creates a dict with all existing values of this data structure

        Returns: dict with values of this data structure
        """
        allArgs = {}
        if self.listId is not None:
            allArgs["list_id"] = self.listId
        if self.similartityLte is not None:
            allArgs["similarity__lte"] = self.similartityLte
        if self.similartityGte is not None:
            allArgs["similarity__gte"] = self.similartityGte
        return allArgs


class Filters:
    """
    Filters determine whether to create face or not

    Attributes:
        gender: Gender. 1 - male, 0 - female.
        ethnicities: Array of integer (ethnicities_filter) - Items Enum:1 2 3 4
                     ( AfricanAmerican - 1; Indian - 2; Asian - 3; Caucasian - 4)
        ageLt: [ 0 .. 100 ] - upper bound for reference age
        ageGte: [ 0 .. 100 ] - lower including bound for reference age
        matchesFilters: a list of MatchFilter objects
    """

    def __init__(
        self,
        gender: Optional[int] = None,
        ethnicities: Optional[List[int]] = None,
        ageLt: Optional[int] = None,
        ageGte: Optional[int] = None,
        matchesFilters: Optional[List[MatchFilter]] = None,
    ):
        self.gender = gender
        self.ethnicities = ethnicities
        self.ageLt = ageLt
        self.ageGte = ageGte
        self.matchesFilters = matchesFilters

    def asDict(self) -> dict:
        """
        Creates a dict with all existing values of this data structure

        Returns: dict with values of this data structure
        """
        allArgs = {}
        if self.gender is not None:
            allArgs["gender"] = self.gender
        if self.ethnicities is not None:
            allArgs["ethnicities"] = self.ethnicities
        if self.ageLt is not None:
            allArgs["age__lt"] = self.ageLt
        if self.ageGte is not None:
            allArgs["age__gte"] = self.ageGte
        if self.matchesFilters is not None:
            allArgs["match"] = [matchesFilter.asDict() for matchesFilter in self.matchesFilters]
        return allArgs


class CreateFacePolicy:
    """
    Determines whether to create faces and link them to lists

    Attributes:
        filters: filters determine whether to create face or not
        createFaces: whether to create faces from events
        setSampleAsAvatar: whether to set a sample as avatar to created faces
    """

    def __init__(self, filters: Optional[Filters] = None, createFaces: int = 0, setSampleAsAvatar: int = 1):
        self.createFaces = createFaces
        self.setSampleAsAvatar = setSampleAsAvatar
        self.filters = filters

    def asDict(self) -> dict:
        """
        Creates a dict with all existing values of this data structure

        Returns: dict with values of this data structure
        """
        allArgs = {"create_faces": self.createFaces, "set_sample_as_avatar": self.setSampleAsAvatar}
        if self.filters is not None:
            allArgs["filters"] = self.filters.asDict()
        return allArgs


class LinkToListPolicy:
    """
    Object with luna list for linking created faces

    Attributes:
        listId: luna list id
        filters: filters determine whether to link created face to the list
    """

    def __init__(self, listId: str, filters: Optional[Filters] = None):
        self.listId = listId
        self.filters = filters

    def asDict(self) -> dict:
        """
        Creates a dict with all existing values of this data structure

        Returns: dict with values of this data structure
        """
        allArgs = {"list_id": self.listId}
        if self.filters is not None:
            allArgs["filters"] = self.filters.asDict()
        return allArgs


class Policies:
    """
    Object that determinates a list of rules for proccessing input images. It can keep parameters
    for detecting faces, extracting attributes. You can set rules for matching attributes and creating faces from them.

    Attributes:
        detectPolicy: policy determinates params for face detecting.
        multifacePolicy: whether to process images with several faces
        extractPolicy: policy determinates params for attributes extracting.
        matchPolicy: list of luna lists with filters for matching. Every event that satisfies the filters
                     for the list will be matched by it
        createFacePolicy: whether to create faces and link it to lists
        linkToListsPolicy: a set of luna lists for linking created faces
    """

    def __init__(
        self,
        detectPolicy: Optional[DetectPolicy] = None,
        multifacePolicy: int = 1,
        extractPolicy: Optional[ExtractPolicy] = None,
        matchPolicy: Optional[List[MatchPolicy]] = None,
        createFacePolicy: Optional[CreateFacePolicy] = None,
        linkToListsPolicy: Optional[List[LinkToListPolicy]] = None,
    ):
        self.detectPolicy = detectPolicy
        self.multifacePolicy = multifacePolicy
        self.extractPolicy = extractPolicy
        self.matchPolicy = matchPolicy
        self.createFacePolicy = createFacePolicy
        self.linkToListsPolicy = linkToListsPolicy

    def asDict(self) -> dict:
        """
        Creates a dict with all existing values of this data structure

        Returns: dict with values of this data structure
        """
        allArgs = {}
        if self.detectPolicy is not None:
            allArgs["detect_policy"] = self.detectPolicy.asDict()

        allArgs["multiface_policy"] = self.multifacePolicy

        if self.extractPolicy is not None:
            allArgs["extract_policy"] = self.extractPolicy.asDict()

        if self.matchPolicy is not None:
            allArgs["match_policy"] = [element.asDict() for element in self.matchPolicy]

        if self.createFacePolicy is not None:
            allArgs["create_face_policy"] = self.createFacePolicy.asDict()

        if self.linkToListsPolicy is not None:
            linkToListPolicyList = [linkToListPolicy.asDict() for linkToListPolicy in self.linkToListsPolicy]
            allArgs["link_to_lists_policy"] = linkToListPolicyList
        return allArgs
