"""
Module implements execution http requests
"""
from __future__ import annotations

import asyncio
import io
from numbers import Real
from ssl import SSLContext
from typing import Any, Coroutine, Dict, Optional, Union

import aiohttp
import ormsgpack
import ujson
from aiohttp.payload import TOO_LARGE_BYTES_BODY

from ..common.luna_request import LunaRequest
from .luna_response import LunaResponse
from .utils import DNSResolver  # compat # pylint: disable=unused-import


def executeRequest(
    request: LunaRequest, session: Optional[aiohttp.ClientSession] = None, raiseError: bool = False, echo: bool = False
) -> LunaResponse:
    """
    The function executes the request in the synchronous mode.

    If raiseError is True and response status code is different from 2xx, an exception is thrown.

    Args:
        request: request to execute
        raiseError: throw exception or not
        session: session to work in
        echo: log request and response message or not

    Returns:
        :class:`~.LunaResponse`
    """
    loop = asyncio.get_event_loop()
    reply = loop.run_until_complete(executeAsyncRequest(request, session, raiseError, echo=echo))
    return reply


async def executeAsyncRequest(
    request: LunaRequest,
    session: Optional[aiohttp.ClientSession] = None,
    raiseError: bool = False,
    stream: bool = False,
    echo: bool = False,
) -> LunaResponse:
    """
    Wrapper that manages stream consuming.

    Args:
        request: request to execute
        session: optional session to work in
        raiseError: whether to throw an exception
        stream: whether to return request for streaming or consume all of the data in place
        echo: log request and response message or not

    Returns:
        :class:`~.LunaResponse`
    """
    resp = LunaResponse(request, session, raiseError, echo)
    if stream:
        return resp

    async with resp.sendRequest():
        await resp.read()

    return resp


def makeAiohttpTimeout(
    session: Optional[aiohttp.ClientSession] = None,
    totalTimeout: Optional[int] = None,
    connectTimeout: Optional[int] = None,
    sockConnectTimeout: Optional[int] = None,
    sockReadTimeout: Optional[int] = None,
) -> Union[aiohttp.ClientTimeout, None]:
    """
    The function casts two timeouts to an 'aiohttp.ClientTimeout' object.

    Args:
        session: session to get default timeouts
        totalTimeout: total request timeout
        connectTimeout: socket connect timeout or obtain connection from session connection pool timeout
        sockConnectTimeout: socket connect timeout
        sockReadTimeout: read chunk from socket timeout
    Returns:
        ClientTimeout, if at least one is specified: not specified timeouts will set from session or default
        None, if none is specified
    """
    if totalTimeout is connectTimeout is sockConnectTimeout is sockReadTimeout is None:
        return None

    defaults = {"total": 60, "connect": 20, "sock_connect": 10, "sock_read": 20}
    if session is not None:
        defaults = {key: getattr(session._timeout, key) for key in defaults}  # pylint: disable=W0212

    timeoutKwargs = {
        "total": totalTimeout,
        "connect": connectTimeout,
        "sock_connect": sockConnectTimeout,
        "sock_read": sockReadTimeout,
    }
    timeoutKwargsFull = {key: value if value is not None else defaults[key] for key, value in timeoutKwargs.items()}

    timeout = aiohttp.ClientTimeout(**timeoutKwargsFull)
    return timeout


def _isNotCompressibleType(mimetype: str) -> bool:
    """
    Check that mimetipe is not compressible. heuristic function
    Args:
        mimetype: mimetype

    Returns:
        True if type supposes any compress otherwise false.
    """
    if mimetype.startswith("application/zip"):
        return True
    if mimetype.startswith("image/"):
        return True
    return False


def upgradeAcceptHeaders(headers: Dict[str, str], accept: Union[str, None], acceptEncoding: Union[str, None]):
    """
    Upgrade accept and accept encoding headers
    Args:
        headers: user defined headers
        accept: user accept header value
        acceptEncoding: user accept-encoding header value

    """
    if accept:
        headers["Accept"] = accept
    else:
        accept = headers.get("Accept")
    if acceptEncoding:
        headers["Accept-Encoding"] = acceptEncoding
    else:
        acceptEncoding = headers.get("Accept-Encoding")
    if not acceptEncoding and accept:
        if _isNotCompressibleType(accept):
            headers["Accept-Encoding"] = "identity"


def makeRequest(
    url: str,
    method: str,
    session: Optional[aiohttp.ClientSession] = None,
    *,
    queryParams: Optional[dict] = None,
    body: Optional[Union[str, bytes, dict, list, aiohttp.Payload, RequestPayload]] = None,
    headers: Optional[dict] = None,
    totalTimeout: Optional[Real] = None,
    connectTimeout: Optional[Real] = None,
    sockConnectTimeout: Optional[Real] = None,
    sockReadTimeout: Optional[Real] = None,
    accept: Optional[str] = None,
    acceptEncoding: Optional[str] = None,
    asyncRequest: bool = False,
    raiseError: Optional[bool] = False,
    stream: bool = False,
    echo: bool = False,
    ssl: Optional[SSLContext] = None,
) -> Union[Coroutine[None, None, LunaResponse], LunaResponse]:
    """
    The function creates and fetches a request.

    Args:
        url: request URL
        method: request method
        session: session to work in
        body: request body
        queryParams: query parameters
        headers: headers for request
        acceptEncoding: accept-encoding header value
        accept: accept header value
        raiseError: if request fails, LunaApiException is raised
        totalTimeout: request's processing timeout in seconds.
        connectTimeout: connection timeout in seconds.
        sockConnectTimeout: socket connection timeout in seconds.
        sockReadTimeout: socket chunk read timeout in seconds.
        asyncRequest: execution in asynchronous mode, disabled by default
        stream: whether to return request for streaming or consume all of the data in place
        echo: log a request and response or not
        ssl: ssl context
    Returns:
        structure with status code, request and decoded Luna API response body is returned.
    """

    timeout = makeAiohttpTimeout(session, totalTimeout, connectTimeout, sockConnectTimeout, sockReadTimeout)
    headers: dict = headers if headers is not None else {}
    if isinstance(body, (dict, list)):
        body = RequestPayload.buildJson(body)

    if isinstance(body, RequestPayload):
        headers.update(body.headers)
        body = body.body

    if isinstance(body, bytes) and len(body) > TOO_LARGE_BYTES_BODY:
        # This allows aiohttp to use stream response class.
        # Might be usefull for large files
        body = io.BytesIO(body)
    upgradeAcceptHeaders(headers, accept=accept, acceptEncoding=acceptEncoding)
    request = LunaRequest(
        url=url, method=method, body=body, params=queryParams, headers=headers, timeout=timeout, ssl=ssl
    )
    if asyncRequest:
        return executeAsyncRequest(request, session, raiseError, stream, echo)

    if stream:
        raise ValueError("stream=True is not supported with sync mode!")

    return executeRequest(request, session, raiseError, echo=echo)


class RequestPayload:
    """
    Container for storing request payload

    Attributes:
        body: Any data that will be passed to request.body
        headers: A dict that will be passed to request.headers
    """

    __slots__ = ("body", "headers")

    def __init__(self, body: Optional[Any] = None, headers: Optional[dict] = None) -> None:
        """Initialize with given args

        Args:
            body: Any data that will be passed to request.body
            headers: A dict that will be passed to request.headers
        """
        self.body = body
        self.headers = headers or {}

    @classmethod
    def buildMsgpack(cls, body: Any, headers: Optional[dict] = None) -> RequestPayload:
        """Build payload with body encoded with msgpack.

        Args:
            body: Any type supported by msgpack
            headers: A dict with request headers
        """
        headers = headers.copy() if headers else {}
        headers["Content-Type"] = "application/msgpack"
        return cls(ormsgpack.packb(body), headers)  # pylint: disable-msg=no-member

    @classmethod
    def buildJson(cls, body: Any, headers: Optional[dict] = None, encoding: str = "utf-8") -> RequestPayload:
        """Build payload with body encoded with json.

        Args:
            body: Any type supported by json encoder
            headers: A dict with request headers
            encoding: Payload encoding (default utf-8)
        """
        headers = headers.copy() if headers else {}
        headers["Content-Type"] = f"application/json; charset={encoding}"
        return cls(ujson.dumps(body, ensure_ascii=False).encode(encoding), headers)

    @classmethod
    def buildFlatbuf(cls, body: bytes, headers: Optional[dict] = None) -> RequestPayload:
        """Build payload with body serialized with flatbuf.

        Args:
            body: A bytes with flatbuf object
            headers: A dict with request headers
        """
        headers = headers.copy() if headers else {}
        headers["Content-Type"] = "application/x-flatbuf"
        return cls(body, headers)

    @classmethod
    def buildMultipart(cls, body: aiohttp.MultipartWriter, headers: Optional[dict] = None) -> RequestPayload:
        """Build payload with multipart body.

        Args:
            body: A multipart body write object
            headers: A dict with request headers
        """
        headers = headers.copy() if headers else {}
        return cls(body, headers)

    def applyAcceptHeader(self, accept: str) -> RequestPayload:
        """
        Apply accept header
        Args:
            accept: content type for accept header
        """
        self.headers["Accept"] = accept
        return self

    def applyAccountIdHeader(self, accountId: Optional[str] = None) -> RequestPayload:
        """
        Apply Luna-Account-Id header
        Args:
            accountId: account id for header
        """
        if accountId is not None:
            self.headers["Luna-Account-Id"] = accountId
        return self
