""" Raw event object structures """
from dataclasses import dataclass
from typing import List, Literal, Optional, TypeVar, Union

from ..events.http_objs import (
    BaseClass,
    BodyAccessories,
    BodyAggregateEstimations,
    BodyBasicAttributes,
    BodyExtraAttributes,
    BodyUpperAttributes,
    Deepfake,
    Emotions,
    EventMatch,
    Liveness,
    Location,
    MaskEstimations,
    Rect,
)

TP = TypeVar("TP", bound=int)

Point = List[TP]
Landmarks = List[Point]


@dataclass
class FaceOcclusionEstimations(BaseClass):
    """
    Face occlusion estimations container.
    """

    clear: float
    chin: float
    correct: float
    mouth: float
    partially: float
    full: float


@dataclass
class FaceOcclusion(BaseClass):
    """
    Face occlusion container.
    """

    predominantOcclusion: str
    estimations: FaceOcclusionEstimations


@dataclass
class Mask(BaseClass):
    """
    Mask class

    Attributes:
        estimations: mask estimations
        predominantMask: predominant mask name
    """

    estimations: MaskEstimations
    predominantMask: str
    faceOcclusion: Optional[FaceOcclusion] = None


@dataclass
class Gaze(BaseClass):
    """
    Gaze class

    Attributes:
        pitch: angle pitch value
        yaw: angle yaw value
    """

    pitch: float
    yaw: float


@dataclass
class HeadPose(BaseClass):
    """
    HeadPose class

    Attributes:
        pitch: angle pitch value
        yaw: angle yaw value
        roll: angle roll value
    """

    pitch: float
    yaw: float
    roll: float


@dataclass
class MouthAttributes(BaseClass):
    """
    Mouth attributes class

    Attributes:
        opened: mouth opened value
        smile: mouth smile value
        occluded: mouth occluded value
        score: mouth score value
    """

    opened: float
    smile: float
    occluded: float
    score: float


@dataclass
class EyePoint(BaseClass):
    """
    Eye point class

    Attributes:
        state: eye state ["open", "closed", "occluded"]
        irisLandmarks: iris landmarks
    """

    state: str
    irisLandmarks: Landmarks


@dataclass
class EyesAttributes(BaseClass):
    """
    Eyes attributes class

    Attributes:
        leftEye: left eye estimations
        rightEye: right eye estimations
    """

    leftEye: EyePoint
    rightEye: EyePoint


@dataclass
class QualityEstimations(BaseClass):
    """
    Quality estimations class

    Attributes:
        blurriness: blurriness estimation value
        dark: dark estimation value
        illumination: illumination estimation value
        specularity: specularity estimation value
        light: light estimation value
    """

    blurriness: float
    dark: float
    illumination: float
    specularity: float
    light: float


@dataclass
class EthnicityEstimations(BaseClass):
    """
    Ethnicity estimations class

    Attributes:
        asian: asian estimation value
        indian: indian estimation value
        caucasian: caucasian estimation value
        africanAmerican: african-american estimation value
    """

    africanAmerican: float
    asian: float
    indian: float
    caucasian: float


@dataclass
class RawEthnicities(BaseClass):
    """
    Raw ethnicities class

    Attributes:
        predominantEthnicity: predominant ethnicity name
        estimations: ethnicity estimations
    """

    predominantEthnicity: str
    estimations: EthnicityEstimations


@dataclass
class RawBasicAttributes(BaseClass):
    """
    Raw basic attributes class

    Attributes:
        age: age
        ethnicities: ethnicities
        gender: gender (0 - female, 1 - male)
    """

    age: int
    gender: int
    ethnicities: RawEthnicities


@dataclass
class RawFaceSampleAttributes(BaseClass):
    """
    Raw face sample class

    Attributes:
        emotions: face emotions
        mask: face mask
        gaze: face gaze
        headPose: face head pose
        mouthAttributes: face mouth attributes
        eyesAttributes: face eyes attributes
        liveness: face liveness
        deepfake: face deepfake
    """

    emotions: Optional[Emotions] = None
    mask: Optional[Mask] = None
    gaze: Optional[Gaze] = None
    headPose: Optional[HeadPose] = None
    mouthAttributes: Optional[MouthAttributes] = None
    eyesAttributes: Optional[EyesAttributes] = None
    liveness: Optional[Liveness] = None
    deepfake: Optional[Deepfake] = None


@dataclass
class FaceQualityFloatThreshold(BaseClass):
    """
    Face quality float threshold
    Attributes:
        min: minimal alloweb value
        max: maximum allowed value
    """

    min: Optional[float] = None
    max: Optional[float] = None


@dataclass
class FaceQualityIntThreshold(BaseClass):
    """
    Face quality int threshold
    Attributes:
        min: minimal alloweb value
        max: maximum allowed value
    """

    min: Optional[int] = None
    max: Optional[int] = None


@dataclass
class FaceQualityCheck(BaseClass):
    """
    Face quality check class
    Attributes:
        name: face quality check name
        objectValue: face quality checks estimation value
        thesholdValue: face quality check threshold
        result: face quality check result: 0 - fail, 1 - success
    """

    name: str
    objectValue: Union[str, float, int]
    thresholdValue: Union[FaceQualityFloatThreshold, FaceQualityIntThreshold, List[str], int]
    result: int


@dataclass
class FaceQuality(BaseClass):
    """
    Face quality class

    Attributes:
        status: face quality checks status for all estimates
        checks: face quality container with all estimates
    """

    status: int
    checks: List[FaceQualityCheck]


@dataclass
class RawFaceDetection(BaseClass):
    """
    Raw face detection class

    Attributes:
        attributes: face detection attributes
        rect: face detection rect
        quality: face quality estimations
        landmarks5: array of 5 landmarks
        landmarks68: array of 68 landmarks
        faceQuality: face quality checks
    """

    attributes: Optional[RawFaceSampleAttributes] = None
    rect: Optional[Rect] = None
    quality: Optional[QualityEstimations] = None
    landmarks5: Optional[Landmarks] = None
    landmarks68: Optional[Landmarks] = None
    faceQuality: Optional[FaceQuality] = None


@dataclass
class RawBodyDetection(BaseClass):
    """
    Raw body detection class

    Attributes:
        attributes: body detection attributes
        rect: body detection rect
    """

    attributes: Optional[BodyExtraAttributes] = None
    rect: Optional[Rect] = None


@dataclass
class RawFace(BaseClass):
    """
    Event face class

    Attributes:
        faceId: created face id
        lists: attach result (list ids)
        eventId: event id
        userData: user data
        externalId: external id
        avatar: avatar for face
        url: location of face
    """

    faceId: str
    lists: Optional[List[str]] = None
    eventId: Optional[str] = None
    userData: Optional[str] = None
    externalId: Optional[str] = None
    avatar: Optional[str] = None
    url: Optional[str] = None


@dataclass
class RawFaceAttributes(BaseClass):
    """
    Raw face attributes class

    Attributes:
        attributeId: attribute id
        score: face descriptor quality score
        samples: resource for attributes
        basicAttributes: basic attributes
        descriptor: descriptor data
        url: location of attributes
    """

    attributeId: Optional[str] = None
    score: Optional[float] = None
    samples: Optional[List[str]] = None
    basicAttributes: Optional[RawBasicAttributes] = None
    descriptor: Optional[str] = None
    url: Optional[str] = None


@dataclass
class RawBodyAttributes(BaseClass):
    """
    Raw face attributes class

    Attributes:
        score: body descriptor quality score
        samples: resource for attributes
        descriptor: descriptor data
    """

    score: Optional[float] = None
    samples: Optional[List[str]] = None
    descriptor: Optional[str] = None


@dataclass
class RawFaceSample(BaseClass):
    """
    Raw face sample class

    Attributes:
        sampleId: sample id
        detection: face detection
        url: location of the sample
    """

    sampleId: Union[str, None]
    detection: RawFaceDetection
    url: Optional[str] = None


@dataclass
class RawBodySample(BaseClass):
    """
    Raw body sample class

    Attributes:
        sampleId: sample id
        detection: body detection
        url: location of the sample
    """

    sampleId: Union[str, None]
    detection: RawBodyDetection
    url: Optional[str] = None


@dataclass
class RawDetectionSamples(BaseClass):
    """
    Detection samples class. The face and the body refer to the same human.

    Attributes:
        face: face sample
        body: body sample
    """

    face: Optional[RawFaceSample] = None
    body: Optional[RawBodySample] = None

    def asDict(self) -> dict:
        """
        Cast self into dict. Used by `ujson.dumps`.

        Returns:
            self as dict
        """
        return {"face": self.face.asDict() if self.face else None, "body": self.body.asDict() if self.body else None}


@dataclass
class RawDetections(BaseClass):
    """
    Raw event detections class

    Attributes:
        filename: source image filename
        samples: samples
        detectTime: time in format RFC 3339
        imageOrigin: the image url - source of the detection
        detectTs: user-defined timestamp relative to something, such as the start of a video
    """

    samples: RawDetectionSamples
    filename: Optional[str] = None
    detectTime: Optional[str] = None
    imageOrigin: Optional[str] = None
    detectTs: Optional[float] = None


@dataclass
class TopFaceMatchResult(BaseClass):
    """
    Face top match result class

    Attributes:
        label: label for matching candidates
        similarity: similarity to the most similar face or event
        faceId: face id
        externalId: external id
    """

    label: str
    similarity: float
    faceId: str
    externalId: Optional[str] = None


@dataclass
class TopEventMatchResult(BaseClass):
    """
    Event top match result class

    Attributes:
        label: label for matching candidates
        similarity: similarity to the most similar face or event
        eventId: event id
        externalId: external id
    """

    label: str
    similarity: float
    eventId: str
    externalId: Optional[str] = None


@dataclass
class DetectionResult(BaseClass):
    """
    Detection result class

    Attributes:
        rect: body detection rect
    """

    rect: Optional[Rect] = None


@dataclass
class ExtractedDetections(BaseClass):
    """
    Class for extracted detections (face and body)

    Attributes:
        sampleId: sample id
        detection: face or body detection
        detectTime: time in format RFC 3339
        imageOrigin: the image url - source of the detection
        detectTs: user-defined timestamp relative to something, such as the start of a video
    """

    sampleId: str
    detection: DetectionResult
    detectTime: str
    imageOrigin: Optional[str] = None
    detectTs: Optional[float] = None


@dataclass
class RawFaceCandidate(BaseClass):
    """
    Raw face candidate class

    Attributes:
        faceId: face id
        createTime: create time
        externalId: external id
        userData: user data
        accountId: account id
        eventId: event id
        avatar: avatar for face
        lists: list of luna lists
    """

    faceId: Optional[str] = None
    createTime: Optional[str] = None
    externalId: Optional[str] = None
    userData: Optional[str] = None
    accountId: Optional[str] = None
    eventId: Optional[str] = None
    avatar: Optional[str] = None
    lists: Optional[List[str]] = None


@dataclass
class RawEventCandidate(BaseClass):
    """
    Raw event candidate class

    Attributes:
        eventId: event id
        streamId: stream id
        source: source
        handlerId: handler id
        externalId: external id
        userData: user data
        createTime: create time
        endTime: end time
        accountId: account id
        faceId: face id
        attachResult: list of attaching faces or bodies
        gender: gender, 0 - female, 1 - male
        age: human age
        emotion: predominant emotion
        mask: predominant mask
        ethnicGroup: ethnicity
        tags: list of tags
        location: information about a place where event is incidented
        trackId: event track id
        topMatch: top match result from match policy
        matchResult: match result for the event
        faceDetections: face detections
        bodyDetections: body detections
    """

    eventId: Optional[str] = None
    accountId: Optional[str] = None
    createTime: Optional[str] = None
    endTime: Optional[str] = None
    handlerId: Optional[str] = None
    externalId: Optional[str] = None
    userData: Optional[str] = None
    streamId: Optional[str] = None
    source: Optional[str] = None
    faceId: Optional[str] = None
    attachResult: Optional[List[str]] = None
    gender: Optional[int] = None
    age: Optional[int] = None
    emotion: Optional[int] = None
    mask: Optional[int] = None
    ethnicGroup: Optional[int] = None
    tags: Optional[List[str]] = None
    location: Optional[Location] = None
    trackId: Optional[str] = None
    topMatch: Optional[Union[TopFaceMatchResult, TopEventMatchResult]] = None
    matchResult: List[EventMatch] = None
    faceDetections: Optional[List[ExtractedDetections]] = None
    bodyDetections: Optional[List[ExtractedDetections]] = None
    bodyBasicAttributes: Optional[BodyBasicAttributes] = None
    upperBody: Optional[BodyUpperAttributes] = None
    accessories: Optional[BodyAccessories] = None


@dataclass
class RawFaceMatchCandidates(BaseClass):
    """
    Raw face match candidates class

    Attributes:
        similarity: similarity
        face: candidate-face
    """

    similarity: float
    face: RawFaceCandidate


@dataclass
class RawEventMatchCandidates(BaseClass):
    """
    Raw event match candidates class

    Attributes:
        similarity: similarity
        event: candidate-event
    """

    similarity: float
    event: RawEventCandidate


@dataclass
class RawMatch(BaseClass):
    """
    Raw event match class

    Attributes:
        label: match label
        candidates: match candidates
    """

    label: str
    candidates: List[Union[RawFaceMatchCandidates, RawEventMatchCandidates]]


@dataclass
class FaceAggregateEstimationsAttributes(BaseClass):
    """
    Event aggregate estimations class

    Attributes:
        liveness: liveness detection
        deepfake: deepfake estimation
        emotions: face emotions
        mask: face mask
    """

    liveness: Optional[Liveness] = None
    deepfake: Optional[Deepfake] = None
    emotions: Optional[Emotions] = None
    mask: Optional[Mask] = None


@dataclass
class FaceAggregateEstimations(BaseClass):
    """
    Event face aggregate estimations class

    Attributes:
        attributes: face aggregate estimations attributes
    """

    attributes: FaceAggregateEstimationsAttributes


@dataclass
class AggregateEstimations(BaseClass):
    """
    Event aggregated estimations class

    Attributes:
        face: face aggregated estimations
        body: body aggregated estimations
    """

    face: Optional[FaceAggregateEstimations] = None
    body: Optional[BodyAggregateEstimations] = None


@dataclass
class RawEvent(BaseClass):
    """
    Raw event class

    Attributes:
        createTime: create time
        endTime: end time
        detections: list of body & face related detections
        faceAttributes: extracted face attributes from samples
        bodyAttributes: extracted body attributes from samples
        externalId: external id
        face: created face
        location: location
        matches: list with match result for the event
        source: source
        streamId: stream id
        tags: list of tags
        trackId: track id
        userData: user data
        aggregateEstimations: aggregate estimations
        meta: user-defined event metadata
    """

    detections: Optional[List[RawDetections]] = None
    location: Optional[Location] = None
    faceAttributes: Optional[RawFaceAttributes] = None
    bodyAttributes: Optional[RawBodyAttributes] = None
    externalId: Optional[str] = None
    face: Optional[RawFace] = None
    matches: Optional[List[RawMatch]] = None
    source: Optional[str] = None
    streamId: Optional[str] = None
    tags: Optional[List[str]] = None
    trackId: Optional[str] = None
    userData: Optional[str] = None
    createTime: Optional[str] = None
    endTime: Optional[str] = None
    aggregateEstimations: Optional[AggregateEstimations] = None
    meta: Optional[dict] = None


@dataclass
class FaceUpgradePolicy(BaseClass):
    """
    Face upgrade policy

    Attributes:
        fdScoreThreshold: face descriptor score threshold
        extractDescriptor: whether to extract face descriptor
        extractBasicAttributes: whether to extract basic attributes
        descriptorVersion: face descriptor version
    """

    fdScoreThreshold: Optional[float] = None
    extractDescriptor: Optional[int] = None
    extractBasicAttributes: Optional[int] = None
    descriptorVersion: Optional[int] = None


@dataclass
class BodyUpgradePolicy(BaseClass):
    """
    Body upgrade policy

    Attributes:
        descriptorVersion: body descriptor version
    """

    descriptorVersion: Optional[int] = None


@dataclass
class FaceUpgradeData(BaseClass):
    """
    Face upgrade data

    Attributes:
        samples: face samples
        policy: face attribute upgrade policy
    """

    samples: List[str]
    policy: Optional[FaceUpgradePolicy] = None


@dataclass
class BodyUpgradeData(BaseClass):
    """
    Body upgrade data

    Attributes:
        samples: body samples
        policy: body attribute upgrade policy
    """

    samples: List[str]
    policy: Optional[BodyUpgradePolicy] = None


@dataclass
class FaceAttributeUpgrade(BaseClass):
    """
    Face attribute upgrade class

    Attributes:
        faceId: face id
        upgradeData: samples and policy to upgrade face attribute
    """

    faceId: str
    upgradeData: FaceUpgradeData


@dataclass
class EventAttributeUpgrade(BaseClass):
    """
    Event attribute upgrade class

    Attributes:
        eventId: event id
        faceUpgradeData: samples and policy to upgrade face attribute
        bodyUpgradeData: samples and policy to upgrade body attribute
    """

    eventId: str
    faceUpgradeData: Optional[FaceUpgradeData] = None
    bodyUpgradeData: Optional[BodyUpgradeData] = None


@dataclass
class StreamEventSourceData(BaseClass):
    """Stream event source data"""

    source: Optional[str] = None
    streamId: Optional[str] = None
    tags: Optional[List[str]] = None
    userData: Optional[str] = None
    externalId: Optional[str] = None
    trackId: Optional[str] = None
    meta: Optional[dict] = None
    eventTime: Optional[str] = None
    eventEndTime: Optional[str] = None
    location: Optional[Location] = None


@dataclass
class ImageOrigin(BaseClass):
    """Image origin"""

    body: bytes
    meta: Optional[dict] = None


@dataclass
class BoundingBox(BaseClass):
    """
    Bounding box
    """

    x: int  # pylint: disable-msg=invalid-name
    y: int  # pylint: disable-msg=invalid-name
    width: int
    height: int


@dataclass
class AggregatedRawImage(BaseClass):
    """Aggregated raw image"""

    body: bytes
    faceBoundingBoxes: Optional[List[BoundingBox]] = None
    bodyBoundingBoxes: Optional[List[BoundingBox]] = None
    meta: Optional[dict] = None


@dataclass
class NonAggregatedRawImage(BaseClass):
    """Non aggregated raw image"""

    body: bytes
    sourceData: Optional[StreamEventSourceData] = None
    faceBoundingBoxes: Optional[List[BoundingBox]] = None
    bodyBoundingBoxes: Optional[List[BoundingBox]] = None
    meta: Optional[dict] = None


@dataclass
class RawDetection(BaseClass):
    """Raw detection"""

    warp: Optional[bytes] = None
    boundingBox: Optional[BoundingBox] = None
    meta: Optional[dict] = None


@dataclass
class AggregatedDetections(BaseClass):
    """Aggregated detections"""

    face: Optional[RawDetection] = None
    body: Optional[RawDetection] = None


@dataclass
class NonAggregatedDetections(BaseClass):
    """Aggregated detections"""

    sourceData: Optional[StreamEventSourceData] = None
    face: Optional[RawDetection] = None
    body: Optional[RawDetection] = None


@dataclass
class StreamEventSourceNonAggregated(BaseClass):
    """Stream event source non aggregated"""

    source: NonAggregatedRawImage | List[NonAggregatedDetections]
    sourceType: Literal["raw_image", "detections"] = None

    filename: Optional[str] = None
    detectTime: Optional[str] = None
    detectTs: Optional[float] = None
    imageOrigin: Optional[Union[ImageOrigin, str]] = None


@dataclass
class StreamEventSourceAggregated(BaseClass):
    """Stream event source aggregated"""

    source: AggregatedRawImage | List[AggregatedDetections]
    sourceType: Literal["raw_image", "detections"] = None

    filename: Optional[str] = None
    detectTime: Optional[str] = None
    detectTs: Optional[float] = None
    imageOrigin: Optional[Union[ImageOrigin, str]] = None


@dataclass
class StreamEventsSource(BaseClass):
    """
    Stream events source
    Request with attributes aggregation fits with `sourceData` at top level and `StreamEventSourceAggregated`
    Request without attributes aggregation fits with `StreamEventSource` and `sourceData` at each `StreamEventSource`
    """

    sources: Union[List[StreamEventSourceNonAggregated], List[StreamEventSourceAggregated]]
    aggregateAttributes: int
    sourceData: Optional[StreamEventSourceData] = None
    useExifInfo: Optional[int] = None
