"""
Module realize lambdas mixin
"""

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

from typing import Awaitable, List, Optional, Union

import aiohttp

from luna3.common.luna_response import LunaResponse
from luna3.common.requests import RequestPayload
from luna3.lunavl.mixins.base import MixinStub


class LambdasMixin(MixinStub):
    """Lambdas Mixin to work with lambda."""

    def _getPayloadLambda(
        self,
        archive: bytes,
        lambdaName: Optional[str] = None,
        description: Optional[str] = None,
        lambdaType: Optional[Literal["standalone", "handlers"]] = None,
        additionalUserDockerCommands: Optional[List[str]] = None,
        baseImageName: Optional[str] = None,
        baseImageTag: Optional[str] = None,
    ) -> RequestPayload:
        """
        Get payload for create/validate lambda

        Args:
            archive: archive file bytes
            lambdaName: name of lambda
            description: description
            lambdaType: lambda type
            additionalUserDockerCommands: additional user docker commands
            baseImageName: base lambda image name
            baseImageTag: base lambda image tag (for custom image)

        Returns:
            payload: RequestPayload
        """
        with aiohttp.MultipartWriter("form-data", self.multipartBoundary) as mpWriter:
            archivePart = mpWriter.append(archive, {"Content-Type": "application/zip"})
            archivePart.set_content_disposition(
                "attachement", name="archive", filename="archive.zip", quote_fields=False
            )

            creds = {"lambda_name": lambdaName}
            if description is not None:
                creds["description"] = description

            if lambdaName is not None:
                credsPart = mpWriter.append_json(creds)
                credsPart.set_content_disposition("form-data", name="credentials", quote_fields=False)

            if lambdaType or additionalUserDockerCommands:
                kwargs = {}
                if lambdaType is not None:
                    kwargs["lambda_type"] = lambdaType
                if additionalUserDockerCommands is not None:
                    kwargs["additional_user_docker_commands"] = additionalUserDockerCommands
                if baseImageName is not None:
                    kwargs["base_image"] = {"name": baseImageName}
                    if baseImageTag is not None:
                        kwargs["base_image"]["tag"] = baseImageTag
                parametersPart = mpWriter.append_json(kwargs)
                parametersPart.set_content_disposition("form-data", name="parameters", quote_fields=False)

            payload = RequestPayload.buildMultipart(body=mpWriter)
        return payload

    def createLambda(
        self,
        archive: bytes,
        lambdaName: str,
        description: Optional[str] = None,
        lambdaType: Optional[Literal["standalone", "handlers"]] = None,
        additionalUserDockerCommands: Optional[List[str]] = None,
        baseImageName: Optional[str] = None,
        baseImageTag: Optional[str] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Create lambda

        See https://docs.visionlabs.ai/luna/latest/ReferenceManuals/APIReferenceManual.html#operation/createLambda

        Args:
            archive: archive file bytes
            lambdaName: name of lambda
            description: lambda description
            lambdaType: lambda type
            additionalUserDockerCommands: additional user docker commands
            baseImageName: base lambda image name
            baseImageTag: base lambda image tag (for custom image)

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` json with lambda_id will be returned.

        >>> import json
        >>> from uuid import uuid4
        >>> from luna3.lunavl.mixins.doctests.config import TEST_CLIENT as client
        >>> from testData.resources import LAMBDA_ARCHIVE
        >>>
        >>> response = client.createLambda(archive=LAMBDA_ARCHIVE, lambdaName=str(uuid4()))
        >>> print(response.statusCode)
        202
        >>> print(json.dumps(response.json, sort_keys=True))  # doctest: +ELLIPSIS
        {"lambda_id": ...}
        >>> _ = client.deleteLambda(lambdaId=response.json["lambda_id"])
        """
        payload = self._getPayloadLambda(
            archive=archive,
            description=description,
            lambdaName=lambdaName,
            lambdaType=lambdaType,
            additionalUserDockerCommands=additionalUserDockerCommands,
            baseImageName=baseImageName,
            baseImageTag=baseImageTag,
        )
        return self.makeRequest(f"{self.baseUri}/lambdas", "POST", body=payload, **kwargs)

    def patchLambda(
        self,
        lambdaId: str,
        lambdaName: Optional[str] = None,
        description: Optional[str] = None,
        lambdaType: Optional[str] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Patch lambda

        Args:
            lambdaId: lambda id
            lambdaName: name of lambda
            description: description
            lambdaType: lambda type

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` json with lambda_id will be returned.

        >>> from uuid import uuid4
        >>> from luna3.lunavl.mixins.doctests.config import TEST_CLIENT as client
        >>> from testData.resources import LAMBDA_ARCHIVE
        >>>
        >>> lambdaId = client.createLambda(archive=LAMBDA_ARCHIVE, lambdaName=str(uuid4())).json["lambda_id"]
        >>> newName = str(uuid4())
        >>> response = client.patchLambda(lambdaId=lambdaId, lambdaName=newName)
        >>> print(response.statusCode)
        204
        """
        jsonBody = {}
        if lambdaName is not None:
            jsonBody["lambda_name"] = lambdaName
        if description is not None:
            jsonBody["description"] = description
        if lambdaType is not None:
            jsonBody["lambda_type"] = lambdaType
        payload = RequestPayload.buildJson(body=jsonBody)
        return self.makeRequest(f"{self.baseUri}/lambdas/{lambdaId}", "PATCH", body=payload, **kwargs)

    def updateLambda(
        self,
        lambdaId: str,
        archive: bytes,
        lambdaName: str,
        description: Optional[str] = None,
        lambdaType: Optional[Literal["standalone", "handlers"]] = None,
        additionalUserDockerCommands: Optional[List[str]] = None,
        baseImageName: Optional[str] = None,
        baseImageTag: Optional[str] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Update lambda

        Args:
            lambdaId: lambda id
            archive: archive file bytes
            lambdaName: name of lambda
            description: description
            lambdaType: lambda type
            additionalUserDockerCommands: additional user docker commands
            baseImageName: base lambda image name
            baseImageTag: base lambda image tag (for custom image)

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` json with lambda_id will be returned.

        >>> from uuid import uuid4
        >>> from time import sleep
        >>> from luna3.lunavl.mixins.doctests.config import TEST_CLIENT as client
        >>> from testData.resources import LAMBDA_ARCHIVE
        >>>
        >>> def wait4start(lambdaId_: str):
        ...     for _ in range(150):
        ...         reply = client.getLambdaStatus(lambdaId_).json
        ...         if reply["status"] == "running":
        ...             break
        ...         if reply["status"] == "terminated":
        ...             raise Exception(f"Expected successfully lambda running. Lambda id: {lambdaId_}")
        ...         sleep(1)
        ...     else:
        ...         raise Exception(f"Lambda execution took more than 150 seconds. Lambda id: {lambdaId_}")
        ...
        >>> lambdaId = client.createLambda(archive=LAMBDA_ARCHIVE, lambdaName=str(uuid4())).json["lambda_id"]
        >>> wait4start(lambdaId)
        >>> newName = str(uuid4())
        >>> response = client.updateLambda(lambdaId=lambdaId, lambdaName=newName, archive=LAMBDA_ARCHIVE)
        >>> print(response.statusCode)
        200
        """
        payload = self._getPayloadLambda(
            archive=archive,
            description=description,
            lambdaName=lambdaName,
            lambdaType=lambdaType,
            additionalUserDockerCommands=additionalUserDockerCommands,
            baseImageName=baseImageName,
            baseImageTag=baseImageTag,
        )
        return self.makeRequest(f"{self.baseUri}/lambdas/{lambdaId}", "PUT", body=payload, **kwargs)

    def updateLambdaImage(
        self,
        lambdaId: str,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Update lambda using latest base image

        Args:
            lambdaId: lambda id

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` json with lambda_id will be returned.

        >>> from uuid import uuid4
        >>> from time import sleep
        >>> from luna3.lunavl.mixins.doctests.config import TEST_CLIENT as client
        >>> from testData.resources import LAMBDA_ARCHIVE
        >>>
        >>> def wait4start(lambdaId_: str):
        ...     for _ in range(150):
        ...         reply = client.getLambdaStatus(lambdaId_).json
        ...         if reply["status"] == "running":
        ...             break
        ...         if reply["status"] == "terminated":
        ...             raise Exception(f"Expected successfully lambda running. Lambda id: {lambdaId_}")
        ...         sleep(1)
        ...     else:
        ...         raise Exception(f"Lambda execution took more than 150 seconds. Lambda id: {lambdaId_}")
        >>>
        >>> lambdaId = client.createLambda(archive=LAMBDA_ARCHIVE, lambdaName=str(uuid4())).json["lambda_id"]
        >>> wait4start(lambdaId)
        >>> response = client.updateLambdaImage(lambdaId=lambdaId)
        >>> print(response.statusCode)
        200
        """
        return self.makeRequest(f"{self.baseUri}/lambdas/{lambdaId}/update", "POST", **kwargs)

    def getLambdas(
        self,
        accountIdFilter: Optional[str] = None,
        createTimeLt: Optional[str] = None,
        createTimeGte: Optional[str] = None,
        lambdaType: Optional[str] = None,
        page: int = 1,
        pageSize: int = 100,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get lambdas by filters.

        Args:
            accountIdFilter: account id filter for get lambdas
            createTimeLt: upper bound of face create time
            createTimeGte: lower bound of face create time
            lambdaType: lambda type
            page: pagination page value
            pageSize: pagination page size value

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` json with lambdas will be returned.

        >>> import json
        >>> from uuid import uuid4
        >>> from luna3.lunavl.mixins.doctests.config import TEST_CLIENT as client
        >>> from testData.resources import LAMBDA_ARCHIVE
        >>>
        >>> _ = client.createLambda(archive=LAMBDA_ARCHIVE, lambdaName=str(uuid4()))
        >>> response = client.getLambdas()
        >>> print(response.statusCode)
        200
        >>> print(json.dumps(response.json["lambdas"][0], sort_keys=True, indent=4))  # doctest: +ELLIPSIS
        {
            "account_id": ...,
            "additional_user_docker_commands": [],
            "base_image": {
                "name": "lpa-lambda-base",
                "tag": ...
            },
            "create_time": ...,
            "description": "",
            "info": {
                "status": ...
            },
            "lambda_id": ...,
            "lambda_name": ...,
            "lambda_type": "standalone",
            "last_update_time": ...,
            "version": 1
        }
        """
        queries = {
            "account_id": accountIdFilter,
            "create_time__lt": createTimeLt,
            "create_time__gte": createTimeGte,
            "lambda_type": lambdaType,
            "page": page,
            "page_size": pageSize,
        }
        return self.makeRequest(f"{self.baseUri}/lambdas", "GET", queryParams=queries, **kwargs)

    def getLambda(
        self, lambdaId: str, accountIdFilter: Optional[str] = None, **kwargs
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get lambda by id.

        Args:
            lambdaId: lambda id
            accountIdFilter: account id filter for get lambda

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` json with lambda_id will be returned.

        >>> import json
        >>> from uuid import uuid4
        >>> from luna3.lunavl.mixins.doctests.config import TEST_CLIENT as client
        >>> from testData.resources import LAMBDA_ARCHIVE
        >>>
        >>> lambdaId = client.createLambda(archive=LAMBDA_ARCHIVE, lambdaName=str(uuid4())).json["lambda_id"]
        >>> response = client.getLambda(lambdaId=lambdaId)
        >>> print(response.statusCode)
        200
        >>> print(json.dumps(response.json, sort_keys=True, indent=4)) # doctest: +ELLIPSIS
        {
            "account_id": ...,
            "additional_user_docker_commands": [],
            "base_image": {
                "name": "lpa-lambda-base",
                "tag": ...
            },
            "create_time": ...,
            "description": "",
            "info": {
                "status": ...
            },
            "lambda_id": ...,
            "lambda_name": ...,
            "lambda_type": "standalone",
            "last_update_time": ...,
            "version": 1
        }
        """
        queries = {"account_id": accountIdFilter}
        return self.makeRequest(f"{self.baseUri}/lambdas/{lambdaId}", "GET", queryParams=queries, **kwargs)

    def checkLambda(
        self, lambdaId: str, accountIdFilter: Optional[str] = None, **kwargs
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Check lambda existence by id.

        Args:
            lambdaId: lambda id
            accountIdFilter: account id filter for check lambda

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.

        >>> from uuid import uuid4
        >>> from luna3.lunavl.mixins.doctests.config import TEST_CLIENT as client
        >>> from testData.resources import LAMBDA_ARCHIVE
        >>>
        >>> lambdaId = client.createLambda(archive=LAMBDA_ARCHIVE, lambdaName=str(uuid4())).json["lambda_id"]
        >>> response = client.checkLambda(lambdaId=lambdaId)
        >>> print(response.statusCode)
        200
        """
        queries = {"account_id": accountIdFilter}
        return self.makeRequest(f"{self.baseUri}/lambdas/{lambdaId}", "HEAD", queryParams=queries, **kwargs)

    def deleteLambda(
        self, lambdaId: str, accountIdFilter: Optional[str] = None, **kwargs
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Delete lambda by id.

        Args:
            lambdaId: lambda id
            accountIdFilter: account id filter for delete lambda

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.

        >>> from uuid import uuid4
        >>> from luna3.lunavl.mixins.doctests.config import TEST_CLIENT as client
        >>> from testData.resources import LAMBDA_ARCHIVE
        >>>
        >>> lambdaId = client.createLambda(archive=LAMBDA_ARCHIVE, lambdaName=str(uuid4())).json["lambda_id"]
        >>> response = client.deleteLambda(lambdaId=lambdaId)
        >>> print(response.statusCode)
        204
        """
        queries = {"account_id": accountIdFilter}
        return self.makeRequest(f"{self.baseUri}/lambdas/{lambdaId}", "DELETE", queryParams=queries, **kwargs)

    def validateLambda(
        self,
        archive: bytes,
        lambdaType: Optional[Literal["standalone", "handlers"]] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Validate lambda.

        Args:
            archive: archive file bytes
            lambdaType: lambda type

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` json with lambda_id will be returned.

        >>> from luna3.lunavl.mixins.doctests.config import TEST_CLIENT as client
        >>> from testData.resources import LAMBDA_ARCHIVE
        >>>
        >>> response = client.validateLambda(archive=LAMBDA_ARCHIVE, lambdaType="standalone")
        >>> print(response.statusCode)
        204
        """
        payload = self._getPayloadLambda(archive=archive, lambdaType=lambdaType)
        return self.makeRequest(f"{self.baseUri}/lambdas/validator", "POST", body=payload, **kwargs)

    def getLambdaStatus(
        self, lambdaId: str, accountIdFilter: Optional[str] = None, **kwargs
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Check lambda status by id.

        Args:
            lambdaId: lambda id
            accountIdFilter: account id filter for get lambda status

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` json with lambda status will be returned.

        >>> import json
        >>> from uuid import uuid4
        >>> from luna3.lunavl.mixins.doctests.config import TEST_CLIENT as client
        >>> from testData.resources import LAMBDA_ARCHIVE
        >>>
        >>> lambdaId = client.createLambda(archive=LAMBDA_ARCHIVE, lambdaName=str(uuid4())).json["lambda_id"]
        >>> response = client.getLambdaStatus(lambdaId=lambdaId)
        >>> print(response.statusCode)
        200
        >>> print(json.dumps(response.json, sort_keys=True)) # doctest: +ELLIPSIS
        {"status": ...}
        """
        queries = {"account_id": accountIdFilter}
        return self.makeRequest(f"{self.baseUri}/lambdas/{lambdaId}/status", "GET", queryParams=queries, **kwargs)

    def getLambdaLogs(
        self, lambdaId: str, tailLines: int = None, accountIdFilter: Optional[str] = None, **kwargs
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get lambda logs

        Args:
            lambdaId: lambda id
            tailLines: number of last string from lambda logs
            accountIdFilter: account id filter for lambda logs

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` json with lambda status will be returned.

        >>> import json
        >>> from uuid import uuid4
        >>> from luna3.lunavl.mixins.doctests.config import TEST_CLIENT as client
        >>> from testData.resources import LAMBDA_ARCHIVE
        >>>
        >>> lambdaId = client.createLambda(archive=LAMBDA_ARCHIVE, lambdaName=str(uuid4())).json["lambda_id"]
        >>> response = client.getLambdaLogs(lambdaId=lambdaId)
        >>> print(response.statusCode)
        200
        >>> print(json.dumps(response.json, sort_keys=True)) # doctest: +ELLIPSIS
        {"logs": [...]}
        """
        queries = {"tail_lines": tailLines, "account_id": accountIdFilter}
        return self.makeRequest(f"{self.baseUri}/lambdas/{lambdaId}/logs", "GET", queryParams=queries, **kwargs)

    def getLambdaImageCreationStatus(
        self, lambdaId: str, accountIdFilter: Optional[str] = None, **kwargs
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Check lambda image creation status by id.

        Args:
            lambdaId: lambda id
            accountIdFilter: account id filter for get lambda image status

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` json with lambda status will be returned.

        >>> import json
        >>> from uuid import uuid4
        >>> from luna3.lunavl.mixins.doctests.config import TEST_CLIENT as client
        >>> from testData.resources import LAMBDA_ARCHIVE
        >>>
        >>> lambdaId = client.createLambda(archive=LAMBDA_ARCHIVE, lambdaName=str(uuid4())).json["lambda_id"]
        >>> response = client.getLambdaImageCreationStatus(lambdaId=lambdaId)
        >>> print(response.statusCode)
        200
        >>> print(json.dumps(response.json, sort_keys=True)) # doctest: +ELLIPSIS
        {"status": ...}
        """
        queries = {"account_id": accountIdFilter}
        return self.makeRequest(f"{self.baseUri}/lambdas/{lambdaId}/image/status", "GET", queryParams=queries, **kwargs)

    def getLambdaImageCreationLogs(
        self, lambdaId: str, tailLines: int = None, **kwargs
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get lambda image creation logs

        Args:
            lambdaId: lambda id
            tailLines: number of last string from lambda logs

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` json with lambda status will be returned.

        >>> import json
        >>> from uuid import uuid4
        >>> from luna3.lunavl.mixins.doctests.config import TEST_CLIENT as client
        >>> from testData.resources import LAMBDA_ARCHIVE
        >>>
        >>> lambdaId = client.createLambda(archive=LAMBDA_ARCHIVE, lambdaName=str(uuid4())).json["lambda_id"]
        >>> response = client.getLambdaImageCreationLogs(lambdaId=lambdaId)
        >>> print(response.statusCode)
        200
        >>> print(json.dumps(response.json, sort_keys=True)) # doctest: +ELLIPSIS
        {"logs": [...]}
        """
        queries = {"tail_lines": tailLines}
        return self.makeRequest(f"{self.baseUri}/lambdas/{lambdaId}/image/logs", "GET", queryParams=queries, **kwargs)

    def getLambdaSpec(
        self,
        lambdaId: str,
        accept: str = "text/html",
        accountIdFilter: Optional[str] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get lambda spec

        Args:
            lambdaId: lambda id
            accept: accept header
            accountIdFilter: account id filter for get lambda docs

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` json with lambda_id will be returned.
        """
        queries = {"account_id": accountIdFilter}
        headers = {"Accept": accept}
        return self.makeRequest(
            f"{self.baseUri}/lambdas/{lambdaId}/docs/spec", method="GET", headers=headers, queryParams=queries, **kwargs
        )

    def proxyLambdaPost(
        self,
        lambdaId: str,
        path: str,
        body: Optional[Union[str, bytes, dict, list, aiohttp.Payload, RequestPayload]] = None,
        queries: Optional[dict] = None,
        headers: Optional[dict] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Proxy POST request to lambda

        Args:
            lambdaId: lambda id
            path: lambda uri path
            body: request body
            headers: request headers
            queries: request queries

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` json with lambda_id will be returned.
        """
        return self.makeRequest(
            f"{self.baseUri}/lambdas/{lambdaId}/proxy/{path}",
            method="POST",
            body=body,
            headers=headers,
            queryParams=queries,
            **kwargs,
        )
