"""
Module realize websocket mixin
"""
from contextlib import contextmanager
from numbers import Real
from typing import List, Optional, Type, Union

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

import aiohttp

from luna3.common.http_objs import GeoBbox
from luna3.common.ws_base import WSBase
from luna3.lunavl.mixins.base import BaseLunaHttpClient
from luna3.sender.wsclient import CLOTHING_COLOR


class WSMixin(WSBase):
    """Websocket Mixin to work with luna-api ws subscription."""

    # auth header; real method is defined above (BaseLunaHttpClient)
    getAuthHeader: Type[BaseLunaHttpClient.getAuthHeader]

    def wsConnect(  # pylint: disable=arguments-differ,arguments-renamed
        self,
        ageLt: Optional[int] = None,
        ageGte: Optional[int] = None,
        gender: Optional[int] = None,
        ethnicGroups: Optional[List[int]] = None,
        tags: Optional[List[str]] = None,
        objectSimilarityLt: Optional[float] = None,
        objectSimilarityGte: Optional[float] = None,
        sources: Optional[List[str]] = None,
        streamIds: Optional[List[str]] = None,
        matchingCandidatesLabels: Optional[List[str]] = None,
        userData: Optional[str] = None,
        geoPosition: Optional[Union[GeoBbox, str]] = None,
        cities: Optional[List[str]] = None,
        areas: Optional[List[str]] = None,
        districts: Optional[List[str]] = None,
        streets: Optional[List[str]] = None,
        houseNumbers: Optional[List[str]] = None,
        handlerIds: Optional[List[str]] = None,
        accountId: Optional[str] = None,
        login: Optional[str] = None,
        password: Optional[str] = None,
        token: Optional[str] = None,
        liveness: Optional[List[int]] = None,
        deepfake: Optional[List[int]] = None,
        masks: Optional[List[int]] = None,
        emotions: Optional[List[int]] = None,
        apparentGender: Optional[List[Literal[0, 1, 2]]] = None,
        apparentAgeLt: Optional[int] = None,
        apparentAgeGte: Optional[int] = None,
        headwearStates: Optional[List[Literal[0, 1, 2]]] = None,
        backpackStates: Optional[List[Literal[0, 1, 2]]] = None,
        sleeveLengths: Optional[List[Literal["short", "long", "undefined"]]] = None,
        upperClothingColors: Optional[List[CLOTHING_COLOR]] = None,
        asyncRequest: Optional[bool] = None,
        messageReceiveTimeout: Optional[Real] = None,
    ) -> Union[aiohttp.client._WSRequestContextManager, contextmanager]:  # pylint: disable=W0212
        """
        Create ws-connection to luna-sender

        See https://docs.visionlabs.ai/luna/latest/ReferenceManuals/APIReferenceManual.html#tag/ws

        Args:
            accountId: account id to connect to
            login: account login
            password: account password
            token: token
            ageLt: [ 0 .. 100 ] - upper bound for reference age
            ageGte: [ 0 .. 100 ] - lower including bound for reference age
            gender: Gender. 1 - male, 0 - female
            ethnicGroups: Array of integer (ethnicities_filter) - Items Enum:1 2 3 4
                         ( AfricanAmerican - 1; Indian - 2; Asian - 3; Caucasian - 4)
            tags: users tags
            objectSimilarityLt: upper excluding similarity boundary
            objectSimilarityGte: lower including similarity boundary
            sources: list of source
            streamIds: list of stream ids
            matchingCandidatesLabels: matching candidate label list
            userData: event user data
            geoPosition: geo position bbox filter specified by coordinates of its center (origin) and some delta
            cities: list of cities
            areas: list of areas
            districts: list of districts
            streets: list of streets
            houseNumbers: list of house numbers
            handlerIds: handler ids
            liveness: allowed liveness states: 0 - spoof, 1 - real, 2 - fake
            deepfake: allowed deepfake states: 0 - fake, 1 - real
            masks: Array of integer (masks) - Items Enum: 1, 2, 3 (1 - missing, 2 - medical mask, 3 - occluded)
            emotions: Array of integer (emotions) - Items Enum: 1, 2, 3, 4, 5, 6, 7 (
                        1 - anger, 2 - disgust, 3 - fear, 4 - happiness, 5 - neutral, 6 - sadness, 7 - surprise)
            apparentGender: apparent gender (body) of event (0, 1, 2) - means (female, male, undefined)
            apparentAgeLt: upper excluding apparent age boundary (body)
            apparentAgeGte:  lower including apparent age boundary (body)
            backpackStates: interest backpack states list ((0, 1, 2) - means (no, yes, unknown)
            headwearStates: interest headwear states list ((0, 1, 2) - means (no, yes, unknown)
            upperClothingColors: interest upper clothing colors list
            sleeveLengths: interest sleeve length list (short, long, undefined)

            asyncRequest: execution in asynchronous mode, disabled by default
            messageReceiveTimeout: message receive timeout (time between start waiting message and message appearance)

        Returns:
            LunaResponse: body contains json with event info.

        >>> import json
        >>> from luna3.lunavl.mixins.doctests.config import TEST_CLIENT as client
        >>> from luna3.common.http_objs import DetectPolicy, ExtractPolicy, NotificationPolicy
        >>> from luna3.common.http_objs import MatchPolicy, Policies, Filters, StoragePolicy
        >>> from luna3.common.http_objs import MatchPolicyCandidates
        >>> from testData.resources import FACE_WARP as image
        >>>
        >>> listId = client.createList(userData="happy_persons", raiseError=True).json["list_id"]
        >>> policies = Policies(
        ...     detectPolicy=DetectPolicy(detectFace=1),
        ...     extractPolicy=ExtractPolicy(extractBasicAttributes=1, fdScoreThreshold=0.0,),
        ...     matchPolicy=[
        ...         MatchPolicy(
        ...             candidates=MatchPolicyCandidates(origin="faces", listId=listId),
        ...             label="happy",
        ...             limit=3,
        ...             threshold=0.5,
        ...             gender=0,
        ...         )
        ...     ],
        ...     storagePolicy=StoragePolicy(notificationPolicy=NotificationPolicy(sendNotification=1,
        ...                                                                       filters=Filters(gender=0))))
        >>> handler = client.createHandler(description="test_handler_notification", policies=policies, raiseError=True)
        >>> with client.wsConnect(messageReceiveTimeout=5) as ws:
        ...     event = client.emitEvents(
        ...         handlerId=handler.json["handler_id"],
        ...         inputData=[image],
        ...         tags="notify_tag",
        ...         imageType=1,
        ...         useExifInfo=1,
        ...         raiseError=True,
        ...         meta={"foo": "bar"},
        ...     )
        ...     WS_RESP = json.loads(ws.recv())
        >>> print(ws.status)
        101
        >>> print(json.dumps(WS_RESP, sort_keys=True, indent=4)) # doctest: +ELLIPSIS
        {
            "Luna-Request-id": ...,
            "event": {
                "aggregate_estimations": {
                    "body": {
                        "attributes": {}
                    },
                    "face": {
                        "attributes": {}
                    }
                },
                "body_attributes": null,
                "detections": [
                    {
                        "detect_time": ...,
                        "detect_ts": null,
                        "filename": "Claudia_Schiffer_warp.jpg",
                        "image_origin": null,
                        "samples": {
                            "body": null,
                            "face": {
                                "detection": {},
                                "sample_id": ...,
                                "url": "/6/samples/faces/..."
                            }
                        }
                    }
                ],
                "event_id": ...,
                "external_id": "",
                "face": null,
                "face_attributes": {
                    "attribute_id": null,
                    "basic_attributes": {
                        "age": ...,
                        "ethnicities": {
                            "estimations": {
                                "african_american": ...,
                                "asian": ...,
                                "caucasian": ...,
                                "indian": ...
                            },
                            "predominant_ethnicity": "caucasian"
                        },
                        "gender": ...
                    },
                    "samples": [...],
                    "score": ...,
                    "url": null
                },
                "location": {},
                "matches": [
                    {
                        "candidates": [],
                        "label": "happy"
                    }
                ],
                "meta": {
                    "foo": "bar"
                },
                "source": null,
                "stream_id": null,
                "tags": [
                    "notify_tag"
                ],
                "track_id": null,
                "url": "/6/events/...",
                "user_data": ""
            },
            "event-create-time": ...,
            "event-end-time": ...,
            "handler_id": ...
        }
        """
        queryParams = {
            "age__lt": ageLt,
            "age__gte": ageGte,
            "gender": gender,
            "ethnic_groups": ethnicGroups,
            "tags": tags,
            "object_similarity__lt": objectSimilarityLt,
            "object_similarity__gte": objectSimilarityGte,
            "sources": sources,
            "stream_ids": streamIds,
            "matching_candidates_labels": matchingCandidatesLabels,
            "user_data": userData,
            "geo_position": geoPosition.buildQueryObject() if isinstance(geoPosition, GeoBbox) else geoPosition,
            "cities": cities,
            "areas": areas,
            "districts": districts,
            "streets": streets,
            "house_numbers": houseNumbers,
            "handler_ids": handlerIds,
            "liveness": liveness,
            "deepfake": deepfake,
            "masks": masks,
            "emotions": emotions,
            "apparent_gender": apparentGender,
            "apparent_age__lt": apparentAgeLt,
            "apparent_age__gte": apparentAgeGte,
            "headwear_states": headwearStates,
            "backpack_states": backpackStates,
            "sleeve_lengths": sleeveLengths,
            "upper_clothing_colors": upperClothingColors,
        }
        headers = self.getAuthHeader(accountId=accountId, login=login, password=password, token=token)
        return super().wsConnect(
            headers=headers, queryParams=queryParams, messageTimeout=messageReceiveTimeout, asyncRequest=asyncRequest
        )
