"""
Module contains backward compatible match objects for luna-python-matcher
"""
import base64
from abc import ABC, abstractmethod
from typing import Dict, List, Optional, Union

from luna3.common.base_api import MISSING
from luna3.common.http_objs import BaseFilters, RawDescriptor
from luna3.public.events import CLOTHING_COLOR, PRIMITIVE_CLOTHING_COLOR

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal


class FaceFilters(BaseFilters):
    """
    Faces filters
    Attributes:
        externalIds: external ids
        faceIds: face ids
        userData: user data
        createTimeGte: create time upper included boundary
        createTimeLt: create time lower excluded boundary
        faceIdGte: face id upper included boundary
        faceIdLt: face id lower excluded boundary
        listId: list id
    """

    __slots__ = (
        "externalIds",
        "faceIds",
        "userData",
        "createTimeGte",
        "createTimeLt",
        "faceIdGte",
        "faceIdLt",
        "listId",
    )

    def __init__(
        self,
        accountId: Optional[str] = None,
        externalIds: Optional[List[str]] = None,
        faceIds: Optional[List[str]] = None,
        userData: Optional[str] = None,
        createTimeGte: Optional[str] = None,
        createTimeLt: Optional[str] = None,
        faceIdGte: Optional[str] = None,
        faceIdLt: Optional[str] = None,
        listId: Optional[str] = None,
    ):
        super().__init__(origin="faces", accountId=accountId)
        self.externalIds = externalIds
        self.faceIds = faceIds
        self.userData = userData
        self.createTimeGte = createTimeGte
        self.createTimeLt = createTimeLt
        self.faceIdGte = faceIdGte
        self.faceIdLt = faceIdLt
        self.listId = listId

    def initFromSnakeKwargs(self, filters: dict) -> "FaceFilters":
        """
        Init face filters from kwargs in snake case
        Args:
            filters: validated filters in snake case
        """
        super().initFromSnakeKwargs(filters)
        self.externalIds = filters.get("external_ids")
        self.faceIds = filters.get("face_ids")
        self.userData = filters.get("user_data")
        self.createTimeGte = filters.get("create_time__gte")
        self.createTimeLt = filters.get("create_time__lt")
        self.faceIdGte = filters.get("face_id__gte")
        self.faceIdLt = filters.get("face_id__lt")
        self.listId = filters.get("list_id")
        return self


class EventFilters(BaseFilters):
    """
    Events filters class
    Attributes
        eventIds: events ids
        eventIdsGte: events id upper included boundary
        eventIdLds: events id lower excluded boundary
        createTimeGte: lower including create time boundary
        createTimeLt: upper excluding create time boundary
        endTimeGte: lower including end time boundary
        endTimeLt: upper excluding end time boundary
        accountId: event account id
        faceIds: ids of faces linked with an event
        handlerIds: handler ids
        externalIds: external ids
        userData: event user data
        sources: list of sources
        streamIds: list of stream ids
        tags: users tags
        topMatchingCandidatesLabel: top matching candidate label
        topSimilarObjectIds: top similar object ids
        topSimilarExternalIds: top similar external ids
        topSimilarObjectSimilarityGte: lower including top similarity boundary
        topSimilarObjectSimilarityLt: upper excluding top similarity boundary
        ageLt: upper excluding age boundary
        ageGte: lower including age boundary
        gender: gender of event (0 or 1)
        emotions: dominant emotions in event (one of (1, 2, 3, 4, 5, 6, 7) -
                  matches "anger", "disgust", "fear", "happiness", "neutral", "sadness", "surprise")
        masks: dominant masks in event (one of (1, 2, 3) - matches "missing", "medical_mask", "occluded")
        liveness: dominant liveness of the face ((0, 1, 2) - means ("spoof", "real", "unknown"))
        deepfake: dominant deepfake of the face ((0, 1) - means ("fake", "real"))
        ethnicGroups: dominant ethnicity of the face ((1, 2, 3, 4) -
                              means (indian, asian, african_american, caucasian)
        apparentAgeLt: apparent age upper excluding boundary
        apparentAgeGte: apparent age upper including boundary
        apparentGender: apparent gender estimation values (0 - female, 1 - male, 2 - undefined)
        headwearStates: headwear states (0 - absent, 1 - present, 2 - undefined)
        headwearApparentColors: headwear apparent colors
        sleeveLengths: sleeve lengths
        upperClothingColors: upper clothing color set
        lowerGarmentTypes: lower garment types
        lowerGarmentColors: lower garment colors
        shoesApparentColors: shoes apparent colors
        backpackStates: backpack states (0 - absent, 1 - present, 2 - undefined)
        areas: list of areas
        streets: list of streets
        districts: list of districts
        cities: list of cities
        houseNumbers: list of house numbers
        geoPosition: geo position bbox filter as dict with the following properties
                     - origin_longitude: origin longitude in degrees (required)
                     - origin_latitude:  origin latitude in degrees (required)
                     - longitude_delta: longitude delta in degrees (if not set, default will be applied)
                     - latitude_delta: latitude delta in degrees (if not set, default will be applied)
        trackIds: track ids
    """

    __slots__ = (
        "eventIdLt",
        "eventIdGte",
        "createTimeGte",
        "createTimeLt",
        "endTimeGte",
        "endTimeLt",
        "eventIds",
        "handlerIds",
        "externalIds",
        "topMatchingCandidatesLabel",
        "topSimilarObjectIds",
        "topSimilarExternalIds",
        "topSimilarObjectSimilarityGte",
        "topSimilarObjectSimilarityLt",
        "ageLt",
        "ageGte",
        "gender",
        "emotions",
        "masks",
        "liveness",
        "deepfake",
        "ethnicGroups",
        "apparentAgeLt",
        "apparentAgeGte",
        "apparentGender",
        "headwearStates",
        "headwearApparentColors",
        "sleeveLengths",
        "upperClothingColors",
        "lowerGarmentTypes",
        "lowerGarmentColors",
        "shoesApparentColors",
        "backpackStates",
        "faceIds",
        "userData",
        "sources",
        "streamIds",
        "tags",
        "cities",
        "areas",
        "districts",
        "streets",
        "houseNumbers",
        "geoPosition",
        "trackIds",
        "meta",
    )

    def __init__(  # pylint: disable=too-many-locals
        self,
        accountId: Optional[str] = MISSING,
        eventIdLt: Optional[str] = MISSING,
        eventIdGte: Optional[str] = MISSING,
        createTimeGte: Optional[str] = MISSING,
        createTimeLt: Optional[str] = MISSING,
        endTimeGte: Optional[str] = MISSING,
        endTimeLt: Optional[str] = MISSING,
        eventIds: Optional[List[str]] = MISSING,
        handlerIds: Optional[List[str]] = MISSING,
        externalIds: Optional[List[str]] = MISSING,
        topMatchingCandidatesLabel: Optional[str] = MISSING,
        topSimilarObjectIds: Optional[List[str]] = MISSING,
        topSimilarExternalIds: Optional[List[str]] = MISSING,
        topSimilarObjectSimilarityGte: Optional[float] = MISSING,
        topSimilarObjectSimilarityLt: Optional[float] = MISSING,
        ageLt: Optional[int] = MISSING,
        ageGte: Optional[int] = MISSING,
        gender: Optional[int] = MISSING,
        emotions: Optional[List[int]] = MISSING,
        masks: Optional[List[int]] = MISSING,
        liveness: Optional[List[int]] = MISSING,
        deepfake: Optional[List[int]] = MISSING,
        ethnicGroups: Optional[List[int]] = MISSING,
        apparentAgeLt: Optional[int] = MISSING,
        apparentAgeGte: Optional[int] = MISSING,
        apparentGender: Optional[List[Literal[0, 1, 2]]] = MISSING,
        headwearStates: Optional[List[Literal[0, 1, 2]]] = MISSING,
        headwearApparentColors: Optional[List[PRIMITIVE_CLOTHING_COLOR]] = MISSING,
        sleeveLengths: Optional[List[Literal["short", "long", "undefined"]]] = MISSING,
        upperClothingColors: Optional[List[CLOTHING_COLOR]] = MISSING,
        lowerGarmentTypes: Optional[Literal["undefined", "trousers", "shorts", "skirt"]] = MISSING,
        lowerGarmentColors: Optional[List[CLOTHING_COLOR]] = MISSING,
        shoesApparentColors: Optional[List[PRIMITIVE_CLOTHING_COLOR]] = MISSING,
        backpackStates: Optional[List[Literal[0, 1, 2]]] = MISSING,
        faceIds: Optional[List[str]] = MISSING,
        userData: Optional[str] = MISSING,
        sources: Optional[List[str]] = MISSING,
        streamIds: Optional[List[str]] = MISSING,
        tags: Optional[List[str]] = MISSING,
        cities: Optional[List[str]] = MISSING,
        areas: Optional[List[str]] = MISSING,
        districts: Optional[List[str]] = MISSING,
        streets: Optional[List[str]] = MISSING,
        houseNumbers: Optional[List[str]] = MISSING,
        geoPosition: Optional[Dict[str, float]] = MISSING,
        trackIds: Optional[List[str]] = MISSING,
        meta: Optional[dict] = MISSING,
    ):
        super().__init__(origin="events", accountId=accountId)
        self.eventIds = eventIds
        self.eventIdGte = eventIdGte
        self.eventIdLt = eventIdLt
        self.createTimeGte = createTimeGte
        self.createTimeLt = createTimeLt
        self.endTimeGte = endTimeGte
        self.endTimeLt = endTimeLt
        self.externalIds = externalIds
        self.faceIds = faceIds
        self.userData = userData
        self.sources = sources
        self.streamIds = streamIds
        self.handlerIds = handlerIds
        self.topMatchingCandidatesLabel = topMatchingCandidatesLabel
        self.topSimilarObjectSimilarityGte = topSimilarObjectSimilarityGte
        self.topSimilarObjectSimilarityLt = topSimilarObjectSimilarityLt
        self.topSimilarObjectIds = topSimilarObjectIds
        self.topSimilarExternalIds = topSimilarExternalIds
        self.tags = tags
        self.emotions = emotions
        self.masks = masks
        self.liveness = liveness
        self.deepfake = deepfake
        self.ethnicGroups = ethnicGroups
        self.gender = gender
        self.ageGte = ageGte
        self.ageLt = ageLt
        self.apparentGender = apparentGender
        self.apparentAgeGte = apparentAgeGte
        self.apparentAgeLt = apparentAgeLt
        self.headwearStates = headwearStates
        self.headwearApparentColors = headwearApparentColors
        self.sleeveLengths = sleeveLengths
        self.upperClothingColors = upperClothingColors
        self.lowerGarmentTypes = lowerGarmentTypes
        self.lowerGarmentColors = lowerGarmentColors
        self.shoesApparentColors = shoesApparentColors
        self.backpackStates = backpackStates
        self.cities = cities
        self.districts = districts
        self.streets = streets
        self.houseNumbers = houseNumbers
        self.areas = areas
        self.geoPosition = geoPosition
        self.trackIds = trackIds
        self.meta = meta

    def initFromSnakeKwargs(self, filters: dict) -> "EventFilters":
        """
        Init event filters from kwargs in snake case
        Args:
            filters: validated filters in snake case
        """
        super().initFromSnakeKwargs(filters)
        self.accountId = filters.get("account_id", MISSING)
        self.eventIds = filters.get("event_ids", MISSING)
        self.eventIdGte = filters.get("event_id__gte", MISSING)
        self.eventIdLt = filters.get("event_id__lt", MISSING)
        self.createTimeGte = filters.get("create_time__gte", MISSING)
        self.createTimeLt = filters.get("create_time__lt", MISSING)
        self.endTimeGte = filters.get("end_time__gte", MISSING)
        self.endTimeLt = filters.get("end_time__lt", MISSING)
        self.externalIds = filters.get("external_ids", MISSING)
        self.faceIds = filters.get("face_ids", MISSING)
        self.userData = filters.get("user_data", MISSING)
        self.sources = filters.get("sources", MISSING)
        self.streamIds = filters.get("stream_ids", MISSING)
        self.handlerIds = filters.get("handler_ids", MISSING)
        self.topMatchingCandidatesLabel = filters.get("top_matching_candidates_label", MISSING)
        self.topSimilarObjectSimilarityGte = filters.get("top_similar_object_similarity__gte", MISSING)
        self.topSimilarObjectSimilarityLt = filters.get("top_similar_object_similarity__lt", MISSING)
        self.topSimilarObjectIds = filters.get("top_similar_object_ids", MISSING)
        self.topSimilarExternalIds = filters.get("top_similar_external_ids", MISSING)
        self.tags = filters.get("tags", MISSING)
        self.emotions = filters.get("emotions", MISSING)
        self.masks = filters.get("masks", MISSING)
        self.liveness = filters.get("liveness", MISSING)
        self.deepfake = filters.get("deepfake", MISSING)
        self.ethnicGroups = filters.get("ethnic_groups", MISSING)
        self.gender = filters.get("gender", MISSING)
        self.ageGte = filters.get("age__gte", MISSING)
        self.ageLt = filters.get("age__lt", MISSING)
        self.apparentGender = filters.get("apparent_gender", MISSING)
        self.apparentAgeGte = filters.get("apparent_age__gte", MISSING)
        self.apparentAgeLt = filters.get("apparent_age__lt", MISSING)
        self.headwearStates = filters.get("headwear_states", MISSING)
        self.headwearApparentColors = filters.get("headwear_apparent_colors", MISSING)
        self.sleeveLengths = filters.get("sleeve_lengths", MISSING)
        self.upperClothingColors = filters.get("upper_clothing_colors", MISSING)
        self.lowerGarmentTypes = filters.get("lower_garment_types", MISSING)
        self.lowerGarmentColors = filters.get("lower_garment_colors", MISSING)
        self.shoesApparentColors = filters.get("shoes_apparent_colors", MISSING)
        self.backpackStates = filters.get("backpack_states", MISSING)
        self.cities = filters.get("cities", MISSING)
        self.districts = filters.get("districts", MISSING)
        self.streets = filters.get("streets", MISSING)
        self.houseNumbers = filters.get("house_numbers", MISSING)
        self.areas = filters.get("areas", MISSING)
        self.geoPosition = filters.get("geo_position", MISSING)
        self.trackIds = filters.get("track_ids", MISSING)
        self.meta = filters.get("meta", MISSING)
        return self

    def asNoneEmptyDict(self):
        """
        Prepare dict of stated filters in snake case
        Returns:
            snake case dict
        """
        return {
            restFilterName: getattr(self, filterName)
            for filterName, restFilterName in self._getFilterNames()
            if getattr(self, filterName) is not MISSING
        }


class AttributeFilters(BaseFilters):
    """
    Attributes filters
    Attributes:
        attributeIds: list of attribute ids
    """

    __slots__ = ("attributeIds",)

    def __init__(self, attributeIds: Optional[List[str]] = None, accountId: Optional[str] = None):
        super().__init__(origin="attributes", accountId=accountId)
        self.attributeIds = attributeIds

    def initFromSnakeKwargs(self, filters: dict) -> "AttributeFilters":
        """
        Init face filters from kwargs in snake case
        Args:
            filters: validated filters in snake case
        """
        super().initFromSnakeKwargs(filters)
        self.attributeIds = filters["attribute_ids"]
        return self


class Candidates:
    """
    Object with type and filters for candidates.

    Attributes:
        candidatesType: resource candidates type ("events" or "faces")
        filters: candidates filters
        order: matching sort order
        limit: matching limit
        threshold: matching threshold
        targets: matching targets
    """

    def __init__(
        self,
        filters: Union[FaceFilters, EventFilters, AttributeFilters],
        order: Optional[Literal["similarity", "create_time_asc", "create_time_desc"]] = None,
        limit: Optional[int] = None,
        threshold: Optional[float] = None,
        targets: Optional[List[str]] = None,
    ):
        self.filters = filters
        self.candidatesType = filters.origin
        self.order = order
        self.limit = limit
        self.threshold = threshold
        self.targets = targets

    def asNoneEmptyDict(self):
        """
        Creates a dict with all existing values of this data structure

        Returns: dict with values of this data structure
        """
        result = {"filters": self.filters.asNoneEmptyDict()}
        if self.order:
            result["order"] = self.order
        if self.limit:
            result["limit"] = self.limit
        if self.targets:
            result["targets"] = self.targets
        if self.threshold:
            result["threshold"] = self.threshold
        return result


class Reference:
    """
    Input object for matching with id of entity and its type ("attribute" or "event" or "face")

    Attributes:
        referenceType: entity type ("attribute" or "event" or "face")
        referenceId: id of entity
    """

    def __init__(self, referenceType: str, referenceId: str):
        self.referenceType = referenceType
        self.referenceId = referenceId

    def asDict(self) -> Dict[str, str]:
        """
        Get dict with all existing values of this data structure

        Returns: dict with id and its type
        """
        return {"type": self.referenceType, "id": self.referenceId}


class BinaryReference(ABC):
    """
    Binary reference

    Attributes:
        referenceType (str): entity type
        referenceId (str): id for identification in the match results
        data (Union[dict, bytes]): data with descriptor
    """

    def __init__(self, referenceType: str, referenceId: str, data: Union[dict, bytes]):
        self.referenceType = referenceType
        self.referenceId = referenceId
        self.data = data

    @abstractmethod
    def encodeData(self) -> Union[dict, str]:
        """Encode data to transfer in JSON"""

    def asDict(self, encode: bool = True) -> Dict[str, Union[str, dict]]:
        """
        Get dict with all existing values of this data structure

        Args:
            encode: Whether to encode bytes in base64 for transfer in JSON.

        Returns: dict with id and its type and data
        """
        res = {"type": self.referenceType, "id": self.referenceId}
        if encode:
            res["data"] = self.encodeData()
        else:
            res["data"] = self.data

        return res


class SDKDescriptorReference(BinaryReference):
    """
    SDK descriptor as reference.
    """

    def encodeData(self) -> Union[dict, str]:
        """Encode data to transfer in JSON"""
        return base64.b64encode(self.data).decode("utf-8")

    def __init__(self, referenceId: str, descriptor: bytes):
        super().__init__("sdk_descriptor", referenceId, descriptor)


class XPKDescriptorReference(BinaryReference):
    """
    XPK file as reference.
    """

    def encodeData(self) -> Union[dict, str]:
        """Encode data to transfer in JSON"""
        return base64.b64encode(self.data).decode("utf-8")

    def __init__(self, referenceId: str, descriptor: bytes):
        super().__init__("xpk_file", referenceId, descriptor)


class RawDescriptorReference(BinaryReference):
    """
    Raw descriptor as reference.
    """

    def encodeData(self) -> Union[dict, str]:
        """Encode data to transfer in JSON"""
        return {
            "version": self.data["version"],
            "descriptor": base64.b64encode(self.data["descriptor"]).decode("utf-8"),
        }

    def __init__(self, referenceId: str, descriptor: RawDescriptor):
        data = {"version": descriptor.version, "descriptor": descriptor.descriptor}
        super().__init__("raw_descriptor", referenceId, data)
