"""
Luna-sender client module
"""
from contextlib import contextmanager
from numbers import Real
from typing import Dict, List, Optional, Union

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

import aiohttp

from ..common.base_api import BaseHTTPAPI
from ..common.http_objs import GeoBbox
from ..common.ws_base import WSBase

CLOTHING_COLOR = Literal[  # pylint: disable-msg=invalid-name
    "black",
    "blue",
    "green",
    "gray",
    "orange",
    "purple",
    "red",
    "white",
    "yellow",
    "pink",
    "brown",
    "beige",
    "khaki",
    "multicolored",
    "undefined",
]

PRIMITIVE_COLOR = Literal[  # pylint: disable-msg=invalid-name
    "black",
    "white",
    "other",
    "undefined",
]


class LunaWSClient(BaseHTTPAPI, WSBase):
    """
    Class for ws connections with Luna API.

    Attributes:
        origin (str): luna-sender protocol, host and port; default "http://127.0.0.1:5080"
        api (int): api version of luna-sender, default 2
        asyncRequest (bool): default mode for request async or blocking
        lunaRequestId (str): Luna-Request-Id.
        totalTimeout (Real): request processing timeout in seconds
        connectTimeout (Real): request socket connect timeout or obtaining connection from session connection pool
        sockReadTimeout (Real): request socket read timeout
        sockConnectTimeout (Real): request socket connect timeout
        messageReceiveTimeout (Real): message receive timeout:
            time between start waiting message and message appearance
    """

    def __init__(
        self,
        accountId: Optional[str] = None,
        origin: str = "http://127.0.0.1:5080",
        api: int = 2,
        asyncRequest: Optional[bool] = False,
        lunaRequestId: Optional[str] = None,
        totalTimeout: Real = 60,
        connectTimeout: Optional[Real] = None,
        sockReadTimeout: Optional[Real] = None,
        sockConnectTimeout: Optional[Real] = None,
        messageReceiveTimeout: Optional[Real] = None,
        *,
        session: aiohttp.ClientSession = None,
        echo: bool = False,
        **kwargs,
    ):
        super().__init__(
            origin,
            api,
            asyncRequest,
            lunaRequestId,
            totalTimeout,
            connectTimeout,
            sockReadTimeout,
            sockConnectTimeout,
            session=session,
            echo=echo,
            **kwargs,
        )
        self.accountId = accountId

        self.messageReceiveTimeout = messageReceiveTimeout

    def getAccountId(self, accountId: Optional[str] = None) -> str:
        """
        Get final account id.

        Args:
            accountId: account id

        Returns:
            If async is None, return self.asyncRequest else async
        """
        return accountId if accountId is not None else self.accountId

    def getAccountIdAsDict(self, accountId: Optional[str] = None) -> Dict[str, str]:
        """
        Get final account id header.

        Args:
            accountId: account id. Pattern:
                ^[0-9]{10},[a-f0-9]{8}-[a-f0-9]{4}-4[a-f0-9]{3}-[89ab][a-f0-9]{3}-[a-f0-9]{12}$
        Returns:
            empty dict if both accountId and self.accountId are None or
                return dict with accountId or self.accountId
        """
        accountId = self.getAccountId(accountId)
        if accountId is not None:
            return {"Luna-Account-Id": accountId}
        return {}

    def connect(
        self,
        ageLt: Optional[int] = None,
        ageGte: Optional[int] = None,
        gender: Optional[int] = None,
        ethnicGroups: Optional[List[int]] = None,
        tags: Optional[List[str]] = None,
        objectSimilarityLt: Optional[float] = None,
        objectSimilarityGte: Optional[float] = None,
        sources: Optional[List[str]] = None,
        streamIds: Optional[List[str]] = None,
        matchingCandidatesLabels: Optional[List[str]] = None,
        userData: Optional[str] = None,
        geoPosition: Optional[Union[GeoBbox, str]] = None,
        cities: Optional[List[str]] = None,
        areas: Optional[List[str]] = None,
        districts: Optional[List[str]] = None,
        streets: Optional[List[str]] = None,
        houseNumbers: Optional[List[str]] = None,
        handlerIds: Optional[List[str]] = None,
        deepfake: Optional[List[int]] = None,
        liveness: Optional[List[int]] = None,
        masks: Optional[List[int]] = None,
        emotions: Optional[List[int]] = None,
        lowerGarmentTypes: Optional[List[int]] = None,
        lowerGarmentColors: Optional[List[CLOTHING_COLOR]] = None,
        shoesColors: Optional[List[PRIMITIVE_COLOR]] = None,
        headwearColors: Optional[List[PRIMITIVE_COLOR]] = None,
        apparentGender: Optional[List[Literal[0, 1, 2]]] = None,
        apparentAgeLt: Optional[int] = None,
        apparentAgeGte: Optional[int] = None,
        headwearStates: Optional[List[Literal[0, 1, 2]]] = None,
        backpackStates: Optional[List[Literal[0, 1, 2]]] = None,
        sleeveLengths: Optional[List[Literal["short", "long", "undefined"]]] = None,
        upperClothingColors: Optional[List[CLOTHING_COLOR]] = None,
        accountId: Optional[str] = None,
        asyncRequest: Optional[bool] = None,
        messageReceiveTimeout: Optional[Real] = None,
    ) -> Union[aiohttp.client._WSRequestContextManager, contextmanager]:  # pylint: disable=W0212
        """
        Create ws-connection to luna-sender

        Args:
            accountId: account id to connect to
            ageLt: [ 0 .. 100 ] - upper bound for reference age
            ageGte: [ 0 .. 100 ] - lower including bound for reference age
            gender: Gender. 1 - male, 0 - female
            ethnicGroups: Array of integer (ethnicities_filter) - Items Enum:1 2 3 4
                         ( AfricanAmerican - 1; Indian - 2; Asian - 3; Caucasian - 4)
            tags: users tags
            objectSimilarityLt: upper excluding similarity boundary
            objectSimilarityGte: lower including similarity boundary
            sources: list of source
            streamIds: stream ids
            matchingCandidatesLabels: matching candidate label list
            userData: event user data
            geoPosition: geo position bbox filter specified by coordinates of its center (origin) and some delta
            cities: list of cities
            areas: list of areas
            districts: list of districts
            streets: list of streets
            houseNumbers: list of house numbers
            handlerIds: handler ids
            deepfake: Array of integer (deepfake) - Items Enum: 0, 1 ( fake - 0; real - 1)
            liveness: Array of integer (liveness) - Items Enum: 0, 1, 2 ( spoof - 0; real - 1; unknown - 2)
            masks: Array of integer (masks) - Items Enum: 1, 2, 3 (1 - missing, 2 - medical mask, 3 - occluded)
            emotions: Array of integer (emotions) - Items Enum: 1, 2, 3, 4, 5, 6, 7 (
                        1 - anger, 2 - disgust, 3 - fear, 4 - happiness, 5 - neutral, 6 - sadness, 7 - surprise)
            lowerGarmentTypes: Array of integer (garment types) - Items Enum: 0, 1, 2, 3(
                        0 - trousers, 1 - shorts, 2 - skirt, 3 - undefined)
            lowerGarmentColors: interest lower garment colors list
            shoesColors: interest shoes colors list
            headwearColors: interest headwear colors list
            apparentGender: apparent gender (body) of event (0, 1, 2) - means (female, male, undefined)
            apparentAgeLt: upper excluding apparent age boundary (body)
            apparentAgeGte:  lower including apparent age boundary (body)
            backpackStates: interest backpack states list ((0, 1, 2) - means (no, yes, unknown)
            headwearStates: interest headwear states list ((0, 1, 2) - means (no, yes, unknown)
            upperClothingColors: interest upper clothing colors list
            sleeveLengths: interest sleeve length list (short, long, undefined)

            asyncRequest: execution in asynchronous mode, disabled by default
            messageReceiveTimeout: message receive timeout (time between start waiting message and message appearance)
        Returns:
            context manager with connection to luna-sender

        >>> import asyncio
        >>> senderAPI = LunaWSClient()
        >>> def syncws():
        ...     with senderAPI.connect() as ws:
        ...         for _ in range(3):
        ...             resp = ws.recv()
        ...             print(resp)
        >>> async def asyncws():
        ...     async with senderAPI.connect(asyncRequest=True) as ws:
        ...         count = 0
        ...         async for resp in ws:
        ...             count += 1
        ...             print(resp.data)
        ...             if count > 2:
        ...                 await ws.close()
        ...                 raise SystemExit(0)
        >>> syncws()
        'some three messages'
        >>> asyncio.run(asyncws())
        'another three messages'
        """
        messageTimeout = messageReceiveTimeout or self.messageReceiveTimeout
        queryParams = {
            "age__lt": ageLt,
            "age__gte": ageGte,
            "gender": gender,
            "ethnic_groups": ethnicGroups,
            "tags": tags,
            "object_similarity__lt": objectSimilarityLt,
            "object_similarity__gte": objectSimilarityGte,
            "sources": sources,
            "stream_ids": streamIds,
            "matching_candidates_labels": matchingCandidatesLabels,
            "user_data": userData,
            "geo_position": geoPosition.buildQueryObject() if isinstance(geoPosition, GeoBbox) else geoPosition,
            "cities": cities,
            "areas": areas,
            "districts": districts,
            "streets": streets,
            "house_numbers": houseNumbers,
            "handler_ids": handlerIds,
            "deepfake": deepfake,
            "liveness": liveness,
            "masks": masks,
            "emotions": emotions,
            "apparent_gender": apparentGender,
            "apparent_age__lt": apparentAgeLt,
            "apparent_age__gte": apparentAgeGte,
            "headwear_states": headwearStates,
            "backpack_states": backpackStates,
            "sleeve_lengths": sleeveLengths,
            "upper_clothing_colors": upperClothingColors,
            "lower_garment_types": lowerGarmentTypes,
            "lower_garment_colors": lowerGarmentColors,
            "shoes_colors": shoesColors,
            "headwear_colors": headwearColors,
        }
        headers = self.getAccountIdAsDict(accountId)
        return super().wsConnect(
            headers=headers, queryParams=queryParams, messageTimeout=messageTimeout, asyncRequest=asyncRequest
        )

    def publish(
        self,
        events: List[dict],
        accountId: Optional[str] = None,
        *,
        handlerId: str,
        eventCreateTime: str,
        eventEndTime: str,
        **kwargs,
    ):
        """
        Publish user generated events.
        Args:
            events: user generated events to be published
            accountId: event account id
            handlerId: event handler id
            eventCreateTime: event creation time
            eventEndTime: event end time
        """
        msg = {
            "events": events,
            "handler_id": handlerId,
            "event_create_time": eventCreateTime,
            "event_end_time": eventEndTime,
        }
        headers = self.getAccountIdAsDict(accountId)
        return self.makeRequest(f"{self.baseUri}/events", "POST", body=msg, headers=headers, **kwargs)
